"""
Activity control nodes.
"""

import math

from gaphas.util import path_ellipse
from gaphas.state import observed, reversible_property
from gaphas.item import Handle, Item, LinePort
from gaphas.constraint import EqualsConstraint, LessThanConstraint
from gaphas.geometry import distance_line_point

from gaphor import UML
from gaphor.core import inject
from gaphor.diagram.diagramitem import DiagramItem
from gaphor.diagram.nameditem import NamedItem
from gaphor.diagram.style import (
    ALIGN_LEFT,
    ALIGN_CENTER,
    ALIGN_TOP,
    ALIGN_RIGHT,
    ALIGN_BOTTOM,
)
from gaphor.diagram.style import get_text_point


DEFAULT_JOIN_SPEC = "and"


class ActivityNodeItem(NamedItem):
    """Basic class for simple activity nodes.
    Simple activity node is not resizable.
    """

    __style__ = {"name-outside": True, "name-padding": (2, 2, 2, 2)}

    def __init__(self, id=None):
        NamedItem.__init__(self, id)
        # Do not allow resizing of the node
        for h in self._handles:
            h.movable = False


class InitialNodeItem(ActivityNodeItem):
    """
    Representation of initial node. Initial node has name which is put near
    top-left side of node.
    """

    __uml__ = UML.InitialNode
    __style__ = {"min-size": (20, 20), "name-align": (ALIGN_LEFT, ALIGN_TOP)}

    RADIUS = 10

    def draw(self, context):
        cr = context.cairo
        r = self.RADIUS
        d = r * 2
        path_ellipse(cr, r, r, d, d)
        cr.set_line_width(0.01)
        cr.fill()

        super(InitialNodeItem, self).draw(context)


class ActivityFinalNodeItem(ActivityNodeItem):
    """Representation of activity final node. Activity final node has name
    which is put near right-bottom side of node.
    """

    __uml__ = UML.ActivityFinalNode
    __style__ = {"min-size": (30, 30), "name-align": (ALIGN_RIGHT, ALIGN_BOTTOM)}

    RADIUS_1 = 10
    RADIUS_2 = 15

    def draw(self, context):
        cr = context.cairo
        r = self.RADIUS_2 + 1
        d = self.RADIUS_1 * 2
        path_ellipse(cr, r, r, d, d)
        cr.set_line_width(0.01)
        cr.fill()

        d = r * 2
        path_ellipse(cr, r, r, d, d)
        cr.set_line_width(0.01)
        cr.set_line_width(2)
        cr.stroke()

        super(ActivityFinalNodeItem, self).draw(context)


class FlowFinalNodeItem(ActivityNodeItem):
    """
    Representation of flow final node. Flow final node has name which is
    put near right-bottom side of node.
    """

    __uml__ = UML.FlowFinalNode
    __style__ = {"min-size": (20, 20), "name-align": (ALIGN_RIGHT, ALIGN_BOTTOM)}

    RADIUS = 10

    def draw(self, context):
        cr = context.cairo
        r = self.RADIUS
        d = r * 2
        path_ellipse(cr, r, r, d, d)
        cr.stroke()

        dr = (1 - math.sin(math.pi / 4)) * r
        cr.move_to(dr, dr)
        cr.line_to(d - dr, d - dr)
        cr.move_to(dr, d - dr)
        cr.line_to(d - dr, dr)
        cr.stroke()

        super(FlowFinalNodeItem, self).draw(context)


class DecisionNodeItem(ActivityNodeItem):
    """
    Representation of decision or merge node.
    """

    __uml__ = UML.DecisionNode
    __style__ = {"min-size": (20, 30), "name-align": (ALIGN_LEFT, ALIGN_TOP)}

    RADIUS = 15

    def __init__(self, id=None):
        ActivityNodeItem.__init__(self, id)
        self._combined = None
        # self.set_prop_persistent('combined')

    def save(self, save_func):
        if self._combined:
            save_func("combined", self._combined, reference=True)
        super(DecisionNodeItem, self).save(save_func)

    def load(self, name, value):
        if name == "combined":
            self._combined = value
        else:
            super(DecisionNodeItem, self).load(name, value)

    @observed
    def _set_combined(self, value):
        # self.preserve_property('combined')
        self._combined = value

    combined = reversible_property(lambda s: s._combined, _set_combined)

    def draw(self, context):
        """
        Draw diamond shape, which represents decision and merge nodes.
        """
        cr = context.cairo
        r = self.RADIUS
        r2 = r * 2 / 3

        cr.move_to(r2, 0)
        cr.line_to(r2 * 2, r)
        cr.line_to(r2, r * 2)
        cr.line_to(0, r)
        cr.close_path()
        cr.stroke()

        super(DecisionNodeItem, self).draw(context)


class ForkNodeItem(Item, DiagramItem):
    """
    Representation of fork and join node.
    """

    element_factory = inject("element_factory")

    __uml__ = UML.ForkNode

    __style__ = {
        "min-size": (6, 45),
        "name-align": (ALIGN_CENTER, ALIGN_BOTTOM),
        "name-padding": (2, 2, 2, 2),
        "name-outside": True,
        "name-align-str": None,
    }

    STYLE_TOP = {"text-align": (ALIGN_CENTER, ALIGN_TOP), "text-outside": True}

    def __init__(self, id=None):
        Item.__init__(self)
        DiagramItem.__init__(self, id)

        h1, h2 = Handle(), Handle()
        self._handles.append(h1)
        self._handles.append(h2)
        self._ports.append(LinePort(h1.pos, h2.pos))

        self._combined = None

        self._join_spec = self.add_text(
            "joinSpec",
            pattern="{ joinSpec = %s }",
            style=self.STYLE_TOP,
            visible=self.is_join_spec_visible,
        )

        self._name = self.add_text(
            "name",
            style={
                "text-align": self.style.name_align,
                "text-padding": self.style.name_padding,
                "text-outside": self.style.name_outside,
                "text-align-str": self.style.name_align_str,
                "text-align-group": "stereotype",
            },
            editable=True,
        )

        self.watch("subject<NamedElement>.name", self.on_named_element_name).watch(
            "subject<JoinNode>.joinSpec", self.on_join_node_join_spec
        )

    def save(self, save_func):
        save_func("matrix", tuple(self.matrix))
        save_func("height", float(self._handles[1].pos.y))
        if self._combined:
            save_func("combined", self._combined, reference=True)
        DiagramItem.save(self, save_func)

    def load(self, name, value):
        if name == "matrix":
            self.matrix = eval(value)
        elif name == "height":
            self._handles[1].pos.y = eval(value)
        elif name == "combined":
            self._combined = value
        else:
            # DiagramItem.load(self, name, value)
            super(ForkNodeItem, self).load(name, value)

    def postload(self):
        subject = self.subject
        if subject and isinstance(subject, UML.JoinNode) and subject.joinSpec:
            self._join_spec.text = self.subject.joinSpec
        self.on_named_element_name(None)
        super(ForkNodeItem, self).postload()

    @observed
    def _set_combined(self, value):
        # self.preserve_property('combined')
        self._combined = value

    combined = reversible_property(lambda s: s._combined, _set_combined)

    def setup_canvas(self):
        super(ForkNodeItem, self).setup_canvas()
        self.register_handlers()

        h1, h2 = self._handles
        cadd = self.canvas.solver.add_constraint
        c1 = EqualsConstraint(a=h1.pos.x, b=h2.pos.x)
        c2 = LessThanConstraint(smaller=h1.pos.y, bigger=h2.pos.y, delta=30)
        self.__constraints = (cadd(c1), cadd(c2))
        list(map(self.canvas.solver.add_constraint, self.__constraints))

    def teardown_canvas(self):
        super(ForkNodeItem, self).teardown_canvas()
        list(map(self.canvas.solver.remove_constraint, self.__constraints))
        self.unregister_handlers()

    def is_join_spec_visible(self):
        """
        Check if join specification should be displayed.
        """
        return (
            isinstance(self.subject, UML.JoinNode)
            and self.subject.joinSpec is not None
            and self.subject.joinSpec != DEFAULT_JOIN_SPEC
        )

    def text_align(self, extents, align, padding, outside):
        h1, h2 = self._handles
        w, _ = self.style.min_size
        h = h2.pos.y - h1.pos.y
        x, y = get_text_point(extents, w, h, align, padding, outside)

        return x, y

    def pre_update(self, context):
        self.update_stereotype()
        Item.pre_update(self, context)
        DiagramItem.pre_update(self, context)

    def post_update(self, context):
        Item.post_update(self, context)
        DiagramItem.post_update(self, context)

    def draw(self, context):
        """
        Draw vertical line - symbol of fork and join nodes. Join
        specification is also drawn above the item.
        """
        Item.draw(self, context)
        DiagramItem.draw(self, context)

        cr = context.cairo

        cr.set_line_width(6)
        h1, h2 = self._handles
        cr.move_to(h1.pos.x, h1.pos.y)
        cr.line_to(h2.pos.x, h2.pos.y)

        cr.stroke()

    def point(self, pos):
        h1, h2 = self._handles
        d, p = distance_line_point(h1.pos, h2.pos, pos)
        # Substract line_width / 2
        return d - 3

    def on_named_element_name(self, event):
        print("on_named_element_name", self.subject)
        subject = self.subject
        if subject:
            self._name.text = subject.name
            self.request_update()

    def on_join_node_join_spec(self, event):
        subject = self.subject
        if subject:
            self._join_spec.text = subject.joinSpec or DEFAULT_JOIN_SPEC
            self.request_update()


def is_join_node(subject):
    """
    Check if ``subject`` is join node. 
    """
    return subject and isinstance(subject, UML.JoinNode)


# vim:sw=4:et



"""
Parsing of UML model elements from string tests.
"""

import unittest
from gaphor.application import Application
from gaphor.UML.elementfactory import ElementFactory
from gaphor.UML.umllex import parse
from gaphor.UML.umllex import attribute_pat, operation_pat, parameter_pat
from gaphor import UML


def dump_prop(prop):
    m = attribute_pat.match(prop)
    # print m.groupdict()


def dump_oper(oper):
    m = operation_pat.match(oper)
    if m:
        g = m.group
    else:
        # set name to oper
        return
    # print g('vis'), g('name'), g('type'), g('mult_l'), g('mult_u'), g('tags')
    if g("params"):
        params = g("params")
        while params:
            m = parameter_pat.match(params)
            g = m.group
            # print ' ', g('dir') or 'in', g('name'), g('type'), g('mult_l'), g('mult_u'), g('default'), g('tags')
            params = g("rest")


dump_prop("#/name")
dump_prop('+ / name : str[1..*] = "aap" { static }')
dump_prop('+ / name : str[*] = "aap" { static }')

dump_oper('myfunc(aap:str = "aap", out two): type')
dump_oper("   myfunc2 ( ): type")
dump_oper('myfunc(aap:str[1] = "aap" { tag1, tag2 }, out two {tag3}): type')


factory = ElementFactory()


class AttributeTestCase(unittest.TestCase):
    """
    Parsing an attribute tests.
    """

    def setUp(self):
        pass

    def tearDown(self):
        factory.flush()

    def test_parse_property_simple(self):
        """Test simple property parsing
        """
        a = factory.create(UML.Property)
        UML.parse(a, "myattr")
        self.assertFalse(a.isDerived)
        self.assertEqual("myattr", a.name)
        self.assertTrue(a.typeValue is None, a.typeValue)
        self.assertTrue(a.lowerValue is None, a.lowerValue)
        self.assertTrue(a.upperValue is None, a.upperValue)
        self.assertTrue(a.defaultValue is None, a.defaultValue)

    def test_parse_property_complex(self):
        """Test complex property parsing
        """
        a = factory.create(UML.Property)

        UML.parse(a, '+ / name : str[0..*] = "aap" { static }')
        self.assertEqual("public", a.visibility)
        self.assertTrue(a.isDerived)
        self.assertEqual("name", a.name)
        self.assertEqual("str", a.typeValue)
        self.assertEqual("0", a.lowerValue)
        self.assertEqual("*", a.upperValue)
        self.assertEqual('"aap"', a.defaultValue)

    def test_parse_property_invalid(self):
        """Test parsing property with invalid syntax
        """
        a = factory.create(UML.Property)

        UML.parse(a, '+ name = str[*] = "aap" { static }')
        self.assertEqual('+ name = str[*] = "aap" { static }', a.name)
        self.assertFalse(a.isDerived)
        self.assertTrue(not a.typeValue)
        self.assertTrue(not a.lowerValue)
        self.assertTrue(not a.upperValue)
        self.assertTrue(not a.defaultValue)


class AssociationEndTestCase(unittest.TestCase):
    """
    Parsing association end tests.
    """

    def setUp(self):
        pass

    def tearDown(self):
        factory.flush()

    def test_parse_association_end(self):
        """Test parsing of association end
        """
        a = factory.create(UML.Association)
        p = factory.create(UML.Property)
        p.association = a

        UML.parse(p, "end")
        self.assertEqual("end", p.name)
        self.assertTrue(not p.typeValue)
        self.assertTrue(not p.lowerValue)
        self.assertTrue(not p.upperValue)
        self.assertTrue(not p.defaultValue)

    def test_parse_multiplicity(self):
        """Test parsing of multiplicity
        """
        a = factory.create(UML.Association)
        p = factory.create(UML.Property)
        p.association = a
        UML.parse(p, "0..2 { tag }")
        self.assertTrue(p.name is None)
        self.assertTrue(not p.typeValue)
        self.assertEqual("0", p.lowerValue)
        self.assertEqual("2", p.upperValue)
        self.assertTrue(not p.defaultValue)

    def test_parse_multiplicity2(self):
        """Test parsing of multiplicity with multiline constraints
        """
        a = factory.create(UML.Association)
        p = factory.create(UML.Property)
        p.association = a
        UML.parse(p, "0..2 { tag1, \ntag2}")
        self.assertTrue(p.name is None)
        self.assertTrue(not p.typeValue)
        self.assertEqual("0", p.lowerValue)
        self.assertEqual("2", p.upperValue)
        self.assertTrue(not p.defaultValue)

    def test_parse_derived_end(self):
        """Test parsing derived association end
        """
        a = factory.create(UML.Association)
        p = factory.create(UML.Property)
        p.association = a
        UML.parse(p, "-/end[*] { mytag}")
        self.assertEqual("private", p.visibility)
        self.assertTrue(p.isDerived)
        self.assertEqual("end", p.name)
        self.assertTrue(not p.typeValue)
        self.assertTrue(not p.lowerValue)
        self.assertEqual("*", p.upperValue)
        self.assertTrue(not p.defaultValue)


class OperationTestCase(unittest.TestCase):
    """
    Operation parsing tests.
    """

    def setUp(self):
        factory.flush()

    def tearDown(self):
        factory.flush()

    def test_parse_operation(self):
        """Test parsing simple operation
        """
        o = factory.create(UML.Operation)
        UML.parse(o, "myfunc()")
        self.assertEqual("myfunc", o.name)
        self.assertTrue(not o.returnResult[0].typeValue)
        self.assertFalse(o.formalParameter)

    def test_parse_operation_return(self):
        """Test parsing operation with return value
        """
        o = factory.create(UML.Operation)
        UML.parse(o, "+ myfunc(): int")
        self.assertEqual("myfunc", o.name)
        self.assertEqual("int", o.returnResult[0].typeValue)
        self.assertEqual("public", o.visibility)
        self.assertTrue(not o.formalParameter)

    def test_parse_operation_2_params(self):
        """Test parsing of operation with two parameters
        """
        o = factory.create(UML.Operation)
        UML.parse(o, "# myfunc2 (a: str, b: int = 3 {  static}): float")
        self.assertEqual("myfunc2", o.name)
        self.assertEqual("float", o.returnResult[0].typeValue)
        self.assertEqual("protected", o.visibility)
        self.assertEqual(2, len(o.formalParameter))
        self.assertEqual("a", o.formalParameter[0].name)
        self.assertEqual("str", o.formalParameter[0].typeValue)
        self.assertTrue(o.formalParameter[0].defaultValue is None)
        self.assertEqual("b", o.formalParameter[1].name)
        self.assertEqual("int", o.formalParameter[1].typeValue)
        self.assertEqual("3", o.formalParameter[1].defaultValue)

    def test_parse_operation_1_param(self):
        """Test parsing of operation with one parameter
        """
        o = factory.create(UML.Operation)
        UML.parse(o, "- myfunc2 (a: node): double")
        self.assertEqual("myfunc2", o.name)
        self.assertEqual("double", o.returnResult[0].typeValue)
        self.assertEqual("private", o.visibility)
        self.assertEqual(1, len(o.formalParameter))
        self.assertEqual("a", o.formalParameter[0].name)
        self.assertEqual("node", o.formalParameter[0].typeValue)
        self.assertTrue(o.formalParameter[0].defaultValue is None)

    def test_parse_operation_invalid_syntax(self):
        """Test operation parsing with invalid syntax
        """
        o = factory.create(UML.Operation)
        UML.parse(o, "- myfunc2: myType2")
        self.assertEqual("- myfunc2: myType2", o.name)



import unittest

import gaphor.UML as UML


class ClassesTestCase(unittest.TestCase):
    def setUp(self):

        self.factory = UML.ElementFactory()

    def tearDown(self):

        del self.factory

    def test_association(self):

        """Testing Association elements in the meta-model"""

        try:

            element = self.factory.create(UML.Association)

        except AttributeError:

            self.fail("Association elements are not part of the meta-model")

        self.assertFalse(
            element.isDerived,
            "The isDerived property should default to False - %s" % element.isDerived,
        )

        property1 = self.factory.create(UML.Property)
        property2 = self.factory.create(UML.Property)

        element.memberEnd = property1
        element.memberEnd = property2

        element.ownedEnd = property1

        element.navigableOwnedEnd = property1

        self.assertTrue(
            property1 in element.member,
            "Namespace.member does not contain memberEnd - %s" % element.member,
        )
        self.assertTrue(
            property2 in element.member,
            "Namespace.member does not contain memberEnd - %s" % element.member,
        )

        self.assertTrue(
            property1 in element.feature,
            "Classifier.feature does not contain ownedEnd - %s" % element.feature,
        )
        self.assertTrue(
            property1 in element.ownedMember,
            "Namespace.ownedMember does not contain ownedEnd - %s" % element.ownedEnd,
        )

        self.assertTrue(
            property1 in element.ownedEnd,
            "Association.ownedEnd does not contain navigableOwnedEnd - %s"
            % element.ownedEnd,
        )

    #    def test_association_class(self):
    #        try:
    #            element = self.factory.create(UML.AssociationClass)
    #        except AttributeError:
    #            self.fail('AssociationClass elements are not part of the meta-model')

    def test_class(self):

        """Testing Class elements in the meta-model"""

        try:

            element = self.factory.create(UML.Class)

        except AttributeError:

            self.fail("Class elements are not part of the meta-model")

        property1 = self.factory.create(UML.Property)
        operation1 = self.factory.create(UML.Operation)

        element.ownedAttribute = property1
        element.ownedOperation = operation1

        self.assertTrue(
            property1 in element.attribute,
            "Classifier.attribute does not contain ownedAttribute - %s"
            % element.attribute,
        )
        self.assertTrue(
            property1 in element.ownedMember,
            "Namespace.ownedMember does not contain ownedAttribute - %s"
            % element.ownedMember,
        )

        self.assertTrue(
            operation1 in element.feature,
            "Classifier.feature does not contain ownedOperation - %s" % element.feature,
        )
        self.assertTrue(
            operation1 in element.ownedMember,
            "Namespace.ownedMember does not contain ownedOperation"
            % element.ownedMember,
        )

    def test_comment(self):

        """Testing Comment elements in the meta-model"""

        try:

            element = self.factory.create(UML.Comment)

        except AttributeError:

            self.fail("Comment elements are not part of the meta-model")

        element.body = "Comment body"

        self.assertTrue(
            element.body == "Comment body", "Incorrect comment body - %s" % element.body
        )

        annotatedElement = self.factory.create(UML.Class)

        element.annotatedElement = annotatedElement

        self.assertTrue(
            annotatedElement in element.annotatedElement,
            "Incorrect annotated element - %s" % element.annotatedElement,
        )

    def test_constraint(self):

        """Testing Constraint elements in the meta-model"""

        try:

            element = self.factory.create(UML.Constraint)

        except AttributeError:

            self.fail("Constraint elements are not part of the meta-model")

        constrainedElement = self.factory.create(UML.Class)

        element.constrainedElement = constrainedElement
        element.specification = "Constraint specification"

        self.assertTrue(
            constrainedElement in element.constrainedElement,
            "Constraint.constrainedElement does not contain the correct element - %s"
            % element.constrainedElement,
        )
        self.assertTrue(
            element.specification == "Constraint specification",
            "Constraint.specification is incorrect - %s" % element.specification,
        )

    def test_dependency(self):

        """Testing Dependency elements in the meta-model"""

        try:

            element = self.factory.create(UML.Dependency)

        except AttributeError:

            self.fail("Dependency elements are not part of the meta-model")

        client = self.factory.create(UML.Package)
        supplier = self.factory.create(UML.Package)

        element.client = client
        element.supplier = supplier

        self.assertTrue(
            client in element.client,
            "Dependency.client does not contain client - %s" % element.client,
        )
        self.assertTrue(
            supplier in element.supplier,
            "Dependency.supplier does not contain supplier - %s" % element.supplier,
        )

    def test_element_import(self):

        try:

            element = self.factory.create(UML.ElementImport)

        except AttributeError:

            self.fail("ElementImport elements are not part of the meta-model")

    def test_enumeration(self):

        try:

            element = self.factory.create(UML.Enumeration)

        except AttributeError:

            self.fail("Enumeration elements are not part of the meta-model")

    def test_generalization(self):

        try:

            element = self.factory.create(UML.Generalization)

        except AttributeError:

            self.fail("Generalization elements are not part of the meta-model")

    def test_interface(self):

        try:

            element = self.factory.create(UML.Interface)

        except AttributeError:

            self.fail("Interface elements are not part of the meta-model")

    def test_namespace(self):

        try:

            element = self.factory.create(UML.Namespace)

        except AttributeError:

            self.fail("Namespace elements are not part of the meta-model")

    def test_operation(self):

        try:

            element = self.factory.create(UML.Operation)

        except AttributeError:

            self.fail("Operation elements are not part of the meta-model")

    def test_package(self):

        try:

            element = self.factory.create(UML.Package)

        except AttributeError:

            self.fail("Package elements are not part of the meta-model")

    def test_parameter(self):

        try:

            element = self.factory.create(UML.Parameter)

        except AttributeError:

            self.fail("Parameter elements are not part of the meta-model")

    def test_property(self):

        try:

            element = self.factory.create(UML.Property)

        except AttributeError:

            self.fail("Property elements are not part of the meta-model")

    def test_realization(self):

        try:

            element = self.factory.create(UML.Realization)

        except AttributeError:

            self.fail("Realization elements are not part of the meta-model")


class Uml2TestCase(unittest.TestCase):
    def test_ids(self):
        factory = UML.ElementFactory()
        c = factory.create(UML.Class)
        assert c.id
        p = factory.create_as(UML.Class, id=False)
        assert p.id is False, p.id

    def test1(self):
        factory = UML.ElementFactory()
        c = factory.create(UML.Class)
        p = factory.create(UML.Package)
        c.package = p
        self.assertEqual(c.package, p)
        self.assertEqual(c.namespace, p)
        self.assertTrue(c in p.ownedElement)

    def testOwnedMember_Unlink(self):
        factory = UML.ElementFactory()
        c = factory.create(UML.Class)
        p = factory.create(UML.Package)
        c.package = p

        c.unlink()

        self.assertEqual([p], factory.lselect())

    #    def test_lower_upper(self):
    #        """
    #        Test MultiplicityElement.{lower|upper}
    #        """
    #        assert UML.MultiplicityElement.lowerValue in UML.MultiplicityElement.lower.subsets
    #
    #        e = UML.MultiplicityElement()
    #        e.lowerValue = '2'
    #        assert e.lower == '2', e.lower
    #
    #        assert UML.MultiplicityElement.upperValue in UML.MultiplicityElement.upper.subsets
    #
    #        e.upperValue = 'up'
    #        assert UML.MultiplicityElement.upper.version == 4, UML.MultiplicityElement.upper.version
    #        assert e.upper == 'up'
    #        e.upperValue = 'down'
    #        assert UML.MultiplicityElement.upper.version == 5, UML.MultiplicityElement.upper.version
    #        assert e.upper == 'down', e.upper
    #
    #        # TODO: test signal handling

    def test_property_is_composite(self):
        p = UML.Property()
        assert p.isComposite == False, p.isComposite
        p.aggregation = "shared"
        assert p.isComposite == False, p.isComposite
        p.aggregation = "composite"
        assert p.isComposite == True, p.isComposite

    def test_association_endType(self):
        factory = UML.ElementFactory()
        c1 = UML.Class()
        c2 = UML.Class()
        a = UML.Association()
        a.memberEnd = UML.Property()
        a.memberEnd = UML.Property()
        a.memberEnd[0].type = c1
        a.memberEnd[1].type = c2
        c1.ownedAttribute = a.memberEnd[0]
        c2.ownedAttribute = a.memberEnd[1]

        assert c1 in a.endType
        assert c2 in a.endType

        c3 = UML.Class()
        a.memberEnd[1].type = c3

        assert c1 in a.endType
        assert c3 in a.endType

    def test_property_navigability(self):
        factory = UML.ElementFactory()
        p = factory.create(UML.Property)
        assert p.navigability is None

        c1 = factory.create(UML.Class)
        c2 = factory.create(UML.Class)
        a = UML.model.create_association(factory, c1, c2)
        assert a.memberEnd[0].navigability is None
        assert a.memberEnd[1].navigability is None

        UML.model.set_navigability(a, a.memberEnd[0], True)
        assert a.memberEnd[0].navigability is True
        assert a.memberEnd[1].navigability is None

        UML.model.set_navigability(a, a.memberEnd[0], False)
        assert a.memberEnd[0].navigability is False
        assert a.memberEnd[1].navigability is None

    def test_namedelement_qualifiedname(self):
        factory = UML.ElementFactory()
        p = factory.create(UML.Package)
        p.name = "Package"
        c = factory.create(UML.Class)
        c.name = "Class"

        self.assertEqual(("Class",), c.qualifiedName)

        p.ownedClassifier = c

        self.assertEqual(("Package", "Class"), c.qualifiedName)

    def test_extension_metaclass(self):
        factory = UML.ElementFactory()
        c = factory.create(UML.Class)
        c.name = "Class"
        s = factory.create(UML.Stereotype)
        s.name = "Stereotype"

        e = UML.model.create_extension(factory, c, s)

        self.assertEqual(c, e.metaclass)

    def test_metaclass_extension(self):
        factory = UML.ElementFactory()
        c = factory.create(UML.Class)
        c.name = "Class"
        s = factory.create(UML.Stereotype)
        s.name = "Stereotype"

        self.assertEqual([], c.extension)
        self.assertEqual([], s.extension)

        e = UML.model.create_extension(factory, c, s)

        print(e.memberEnd)
        self.assertEqual([e], c.extension)
        self.assertEqual([], s.extension)
        assert e.ownedEnd.type is s

    def test_operation_parameter_deletion(self):
        factory = UML.ElementFactory()
        self.assertEqual(0, len(factory.lselect()))

        c = factory.create(UML.Class)
        c.name = "Class"
        o = factory.create(UML.Operation)
        c.ownedOperation = o
        UML.parse(o, "a(x: int, y: int)")

        c.unlink()

        self.assertEqual(0, len(factory.lselect()), factory.lselect())



import unittest


class ActionManagerTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testLoadAll(self):
        from gaphor.application import Application

        Application.init()
        am = Application.get_service("action_manager")
        ui = am.ui_manager.get_ui()
        print(ui)



"""
Toolbox.
"""

import logging
from zope import component
from zope.interface import implementer

from gi.repository import GObject
from gi.repository import Gdk
from gi.repository import Gtk

from gaphor.core import _, inject, toggle_action, build_action_group
from gaphor.interfaces import IActionProvider
from gaphor.ui.interfaces import IUIComponent, IDiagramPageChange
from gaphor.ui.diagramtoolbox import TOOLBOX_ACTIONS

log = logging.getLogger(__name__)


@implementer(IUIComponent, IActionProvider)
class Toolbox(object):

    TARGET_STRING = 0
    TARGET_TOOLBOX_ACTION = 1
    DND_TARGETS = [
        Gtk.TargetEntry.new("STRING", Gtk.TargetFlags.SAME_APP, TARGET_STRING),
        Gtk.TargetEntry.new("text/plain", Gtk.TargetFlags.SAME_APP, TARGET_STRING),
        Gtk.TargetEntry.new(
            "gaphor/toolbox-action", Gtk.TargetFlags.SAME_APP, TARGET_TOOLBOX_ACTION
        ),
    ]

    title = _("Toolbox")
    placement = ("left", "diagrams")

    component_registry = inject("component_registry")
    main_window = inject("main_window")
    properties = inject("properties")

    menu_xml = """
      <ui>
        <menubar name="mainwindow">
          <menu action="diagram">
            <separator/>
            <menuitem action="reset-tool-after-create" />
            <separator/>
          </menu>
        </menubar>
      </ui>
    """

    def __init__(self, toolbox_actions=TOOLBOX_ACTIONS):
        self._toolbox = None
        self._toolbox_actions = toolbox_actions
        self.action_group = build_action_group(self)
        self.action_group.get_action("reset-tool-after-create").set_active(
            self.properties.get("reset-tool-after-create", True)
        )
        self.buttons = []
        self.shortcuts = {}

    def open(self):
        widget = self.construct()
        self.main_window.window.connect_after(
            "key-press-event", self._on_key_press_event
        )
        self.component_registry.register_handler(self._on_diagram_page_change)
        return widget

    def close(self):
        if self._toolbox:
            self._toolbox.destroy()
            self._toolbox = None
        self.component_registry.unregister_handler(self._on_diagram_page_change)

    def construct(self):
        def toolbox_button(action_name, stock_id, label, shortcut):
            button = Gtk.ToggleToolButton.new()
            button.set_icon_name(stock_id)
            button.action_name = action_name
            if label:
                button.set_tooltip_text("%s (%s)" % (label, shortcut))

            # Enable Drag and Drop
            inner_button = button.get_children()[0]
            inner_button.drag_source_set(
                Gdk.ModifierType.BUTTON1_MASK | Gdk.ModifierType.BUTTON3_MASK,
                self.DND_TARGETS,
                Gdk.DragAction.COPY | Gdk.DragAction.LINK,
            )
            inner_button.drag_source_set_icon_stock(stock_id)
            inner_button.connect(
                "drag-data-get", self._button_drag_data_get, action_name
            )

            return button

        toolbox = Gtk.ToolPalette.new()
        toolbox.connect("destroy", self._on_toolbox_destroyed)

        for title, items in self._toolbox_actions:
            tool_item_group = Gtk.ToolItemGroup.new(title)
            for action_name, label, stock_id, shortcut in items:
                button = toolbox_button(action_name, stock_id, label, shortcut)
                tool_item_group.insert(button, -1)
                button.show()
                self.buttons.append(button)
                self.shortcuts[shortcut] = action_name
            toolbox.add(tool_item_group)
            tool_item_group.show()

        toolbox.show()

        self._toolbox = toolbox
        return toolbox

    def _on_key_press_event(self, view, event):
        """
        Grab top level window events and select the appropriate tool based on the event.
        """
        if event.get_state() & Gdk.ModifierType.SHIFT_MASK or (
            event.get_state() == 0 or event.get_state() & Gdk.ModifierType.MOD2_MASK
        ):
            keyval = Gdk.keyval_name(event.keyval)
            self.set_active_tool(shortcut=keyval)

    def _on_toolbox_destroyed(self, widget):
        self._toolbox = None

    @toggle_action(name="reset-tool-after-create", label=_("_Reset tool"), active=False)
    def reset_tool_after_create(self, active):
        self.properties.set("reset-tool-after-create", active)

    @component.adapter(IDiagramPageChange)
    def _on_diagram_page_change(self, event):
        self.update_toolbox(event.diagram_page.toolbox.action_group)

    def update_toolbox(self, action_group):
        """
        Update the buttons in the toolbox. Each button should be connected
        by an action. Each button is assigned a special _action_name_
        attribute that can be used to fetch the action from the ui manager.
        """
        if not self._toolbox:
            return

        for button in self.buttons:
            action_name = button.action_name
            action = action_group.get_action(action_name)
            if action:
                button.set_related_action(action)

    def set_active_tool(self, action_name=None, shortcut=None):
        """
        Set the tool based on the name of the action
        """
        # HACK:
        toolbox = self._toolbox
        if shortcut and toolbox:
            action_name = self.shortcuts.get(shortcut)
            log.debug("Action for shortcut %s: %s" % (shortcut, action_name))
            if not action_name:
                return

    def _button_drag_data_get(self, button, context, data, info, time, action_name):
        """The drag-data-get event signal handler.

        The drag-data-get signal is emitted on the drag source when the drop
        site requests the data which is dragged.

        Args:
            button (Gtk.Button): The button that received the signal.
            context (Gdk.DragContext): The drag context.
            data (Gtk.SelectionData): The data to be filled with the dragged
                data.
            info (int): The info that has been registered with the target in
                the Gtk.TargetList
            time (int): The timestamp at which the data was received.

        """
        data.set(type=data.get_target(), format=8, data=action_name.encode())



"""
The Sanitize module is dedicated to adapters (stuff) that keeps
the model clean and in sync with diagrams.
"""

from logging import getLogger
from zope import component

from zope.interface import implementer

from gaphor import UML
from gaphor.UML.interfaces import IAssociationDeleteEvent, IAssociationSetEvent
from gaphor.core import inject
from gaphor.interfaces import IService


@implementer(IService)
class SanitizerService(object):
    """
    Does some background cleanup jobs, such as removing elements from the
    model that have no presentations (and should have some).
    """

    logger = getLogger("Sanitizer")

    component_registry = inject("component_registry")
    element_factory = inject("element_factory")
    property_dispatcher = inject("property_dispatcher")

    def __init__(self):
        pass

    def init(self, app=None):
        self.component_registry.register_handler(self._unlink_on_presentation_delete)
        self.component_registry.register_handler(self._unlink_on_stereotype_delete)
        self.component_registry.register_handler(self._unlink_on_extension_delete)
        self.component_registry.register_handler(self._disconnect_extension_end)

    def shutdown(self):
        self.component_registry.unregister_handler(self._unlink_on_presentation_delete)
        self.component_registry.unregister_handler(self._unlink_on_stereotype_delete)
        self.component_registry.unregister_handler(self._unlink_on_extension_delete)
        self.component_registry.unregister_handler(self._disconnect_extension_end)

    @component.adapter(IAssociationDeleteEvent)
    def _unlink_on_presentation_delete(self, event):
        """
        Unlink the model element if no more presentations link to the `item`'s
        subject or the deleted item was the only item currently linked.
        """

        self.logger.debug("Handling IAssociationDeleteEvent")
        # self.logger.debug('Property is %s' % event.property.name)
        # self.logger.debug('Element is %s' % event.element)
        # self.logger.debug('Old value is %s' % event.old_value)

        if event.property is UML.Element.presentation:
            old_presentation = event.old_value
            if old_presentation and not event.element.presentation:
                event.element.unlink()

    def perform_unlink_for_instances(self, st, meta):

        self.logger.debug("Performing unlink for instances")
        # self.logger.debug('Stereotype is %s' % st)
        # self.logger.debug('Meta is %s' % meta)

        inst = UML.model.find_instances(self.element_factory, st)

        for i in list(inst):
            for e in i.extended:
                if not meta or isinstance(e, meta):
                    i.unlink()

    @component.adapter(IAssociationDeleteEvent)
    def _unlink_on_extension_delete(self, event):
        """
        Remove applied stereotypes when extension is deleted.
        """

        self.logger.debug("Handling IAssociationDeleteEvent")
        # self.logger.debug('Property is %s' % event.property.name)
        # self.logger.debug('Element is %s' % event.element)
        # self.logger.debug('Old value is %s' % event.old_value)

        if (
            isinstance(event.element, UML.Extension)
            and event.property is UML.Association.memberEnd
            and event.element.memberEnd
        ):
            p = event.element.memberEnd[0]
            ext = event.old_value
            if isinstance(p, UML.ExtensionEnd):
                p, ext = ext, p
            st = ext.type
            meta = p.type and getattr(UML, p.type.name)
            self.perform_unlink_for_instances(st, meta)

    @component.adapter(IAssociationSetEvent)
    def _disconnect_extension_end(self, event):

        self.logger.debug("Handling IAssociationSetEvent")
        # self.logger.debug('Property is %s' % event.property.name)
        # self.logger.debug('Element is %s' % event.element)
        # self.logger.debug('Old value is %s' % event.old_value)

        if event.property is UML.ExtensionEnd.type and event.old_value:
            ext = event.element
            p = ext.opposite
            if not p:
                return
            st = event.old_value
            meta = getattr(UML, p.type.name)
            self.perform_unlink_for_instances(st, meta)

    @component.adapter(IAssociationDeleteEvent)
    def _unlink_on_stereotype_delete(self, event):
        """
        Remove applied stereotypes when stereotype is deleted.
        """

        self.logger.debug("Handling IAssociationDeleteEvent")
        # self.logger.debug('Property is %s' % event.property)
        # self.logger.debug('Element is %s' % event.element)
        # self.logger.debug('Old value is %s' % event.old_value)

        if event.property is UML.InstanceSpecification.classifier:
            if isinstance(event.old_value, UML.Stereotype):
                event.element.unlink()



"""The core UML metamodel events."""

from zope.interface import implementer

from gaphor.UML.interfaces import (
    IAssociationAddEvent,
    IAssociationDeleteEvent,
    IElementCreateEvent,
)
from gaphor.UML.interfaces import (
    IAttributeChangeEvent,
    IAssociationChangeEvent,
    IAssociationSetEvent,
)
from gaphor.UML.interfaces import (
    IElementFactoryEvent,
    IModelFactoryEvent,
    IElementDeleteEvent,
    IFlushFactoryEvent,
)


@implementer(IAttributeChangeEvent)
class AttributeChangeEvent(object):
    """A UML attribute has changed value."""

    def __init__(self, element, attribute, old_value, new_value):
        """Constructor.  The element parameter is the element with the
        changing attribute.  The attribute parameter is the parameter
        element that changed.  The old_value is the old value of the attribute
        and the new_value is the new value of the attribute."""

        self.element = element
        self.property = attribute
        self.old_value = old_value
        self.new_value = new_value


@implementer(IAssociationChangeEvent)
class AssociationChangeEvent(object):
    """An association UML element has changed."""

    def __init__(self, element, association):
        """Constructor.  The element parameter is the element the association
        is changing from.  The association parameter is the changed
        association element."""

        self.element = element
        self.property = association


@implementer(IAssociationSetEvent)
class AssociationSetEvent(AssociationChangeEvent):
    """An association element has been set."""

    def __init__(self, element, association, old_value, new_value):
        """Constructor.  The element parameter is the element setting the
        association element.  The association parameter is the association
        element being set.  The old_value parameter is the old association
        and the new_value parameter is the new association."""

        AssociationChangeEvent.__init__(self, element, association)
        self.old_value = old_value
        self.new_value = new_value


@implementer(IAssociationAddEvent)
class AssociationAddEvent(AssociationChangeEvent):
    """An association element has been added."""

    def __init__(self, element, association, new_value):
        """Constructor.  The element parameter is the element the association
        has been added to.  The association parameter is the association
        element being added."""

        AssociationChangeEvent.__init__(self, element, association)
        self.new_value = new_value


@implementer(IAssociationDeleteEvent)
class AssociationDeleteEvent(AssociationChangeEvent):
    """An association element has been deleted."""

    def __init__(self, element, association, old_value):
        """Constructor.  The element parameter is the element the association
        has been deleted from.  The association parameter is the deleted
        association element."""

        AssociationChangeEvent.__init__(self, element, association)
        self.old_value = old_value


class DerivedChangeEvent(AssociationChangeEvent):
    """A derived property has changed."""

    pass


@implementer(IAssociationSetEvent)
class DerivedSetEvent(DerivedChangeEvent):
    """A generic derived set event."""

    def __init__(self, element, association, old_value, new_value):
        """Constructor.  The element parameter is the element to which the
        derived set belongs.  The association parameter is the association
        of the derived set."""

        AssociationChangeEvent.__init__(self, element, association)
        self.old_value = old_value
        self.new_value = new_value


@implementer(IAssociationAddEvent)
class DerivedAddEvent(DerivedChangeEvent):
    """A derived property has been added."""

    def __init__(self, element, association, new_value):
        """Constructor.  The element parameter is the element to which the
        derived property belongs.  The association parameter is the 
        association of the derived property."""

        AssociationChangeEvent.__init__(self, element, association)
        self.new_value = new_value


@implementer(IAssociationDeleteEvent)
class DerivedDeleteEvent(DerivedChangeEvent):
    """A derived property has been deleted."""

    def __init__(self, element, association, old_value):
        """Constructor.  The element parameter is the element to which the
        derived property belongs.  The association parameter is the 
        association of the derived property."""

        AssociationChangeEvent.__init__(self, element, association)
        self.old_value = old_value


@implementer(IAssociationSetEvent)
class RedefineSetEvent(AssociationChangeEvent):
    """A redefined property has been set."""

    def __init__(self, element, association, old_value, new_value):
        """Constructor.  The element parameter is the element to which the
        property belongs.  The association parameter is association of the 
        property."""

        AssociationChangeEvent.__init__(self, element, association)
        self.old_value = old_value
        self.new_value = new_value


@implementer(IAssociationAddEvent)
class RedefineAddEvent(AssociationChangeEvent):
    """A redefined property has been added."""

    def __init__(self, element, association, new_value):
        """Constructor.  The element parameter is the element to which the
        property belongs.  The association parameter is the association of
        the property."""

        AssociationChangeEvent.__init__(self, element, association)
        self.new_value = new_value


@implementer(IAssociationDeleteEvent)
class RedefineDeleteEvent(AssociationChangeEvent):
    """A redefined property has been deleted."""

    def __init__(self, element, association, old_value):
        """Constructor.  The element parameter is the element to which the
        property belongs.  The association parameter is the association of
        the property."""

        AssociationChangeEvent.__init__(self, element, association)
        self.old_value = old_value


@implementer(IElementCreateEvent)
class DiagramItemCreateEvent(object):
    """A diagram item has been created."""

    def __init__(self, element):
        """Constructor.  The element parameter is the element being created."""

        self.element = element


@implementer(IElementCreateEvent, IElementFactoryEvent)
class ElementCreateEvent(object):
    """An element has been created."""

    def __init__(self, service, element):
        """Constructor.  The service parameter is the service responsible
        for creating the element.  The element parameter is the element being
        created."""

        self.service = service
        self.element = element


@implementer(IElementDeleteEvent, IElementFactoryEvent)
class ElementDeleteEvent(object):
    """An element has been deleted."""

    def __init__(self, service, element):
        """Constructor.  The service parameter is the service responsible for
        deleting the element.  The element parameter is the element being
        deleted."""

        self.service = service
        self.element = element


@implementer(IModelFactoryEvent)
class ModelFactoryEvent(object):
    """A generic element factory event."""

    def __init__(self, service):
        """Constructor.  The service parameter is the service the emitted the
        event."""

        self.service = service


@implementer(IFlushFactoryEvent)
class FlushFactoryEvent(object):
    """The element factory has been flushed."""

    def __init__(self, service):
        """Constructor.  The service parameter is the service responsible for
        flushing the factory."""

        self.service = service
